"""
expression.py

Classes:
    Var
        - Base class for individual variables
    Expression
        - Inherits from Var
        - Can be combined into larger expressions
"""
from typing import Union
from numbers import Number

import numpy as np

import superjacob as sj


class Var:
    """A Variable.

    Attributes:
        :name: str -- The common name for the variable (e.g. 'x', 'y', 'x1')

    Methods:
        eval () -> Number -- Evaluate the variable for a given input (always return the number itself)

    """
    def __init__(self, name, length=1):
        self._vars = None
        self.name = name
        self.length = length

    def eval(self, x: Union[Number, np.ndarray]):
        """
        Evaluate the variable at x (identity function)

        :param x: Number|np.ndarray -- The point to evaluate the Var at
        :return: Number -- The number x itself
        """
        self._check_length(x)
        return x

    def deriv(self, x: Union[bool, Number, np.ndarray], var=None) -> Union[Number, np.ndarray]:
        """Evaluate the derivative of this variable.

        `x` is a boolean indicator variable
            True (or non-zero) if this variable is being differentiated
            False (or zero) if this variable is not being differentiated

        :param x: bool | Number | np.ndarray -- Whether the derivative is being taken with
            respect to this variable
        :param var: Var -- The variable with respect to which the derivative is being taken
        :return: int | np.ndarray -- 1 or 0 (or array of ones with dimension `self.length`)
        """
        self._check_length(x)
        if x is not None:
            if self.length == 1:
                return 1
            else:
                return np.ones(self.length)
        else:
            return 0

    def _check_length(self, x):
        try:
            assert len(x) == self.length, f'Incorrect input size (required: {self.length}, given: {len(x)}'
        except TypeError:  # Assuming that this is a single number
            assert self.length == 1, f'Incorrect input size (required: {self.length}, given: 1'

    @property
    def vars(self):
        return [self]

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
        
    def __call__(self, *args, **kwargs):
        return self.eval(*args)

    def __add__(self, other):
        return sj.add(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return sj.sub(self, other)

    def __rsub__(self, other):
        return sj.sub(other, self)

    def __mul__(self, other):
        return sj.mul(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return sj.div(self, other)

    def __rtruediv__(self, other):
        return sj.div(other, self)

    def __pow__(self, power):
        return sj.pow(self, power)

    def __rpow__(self, base):
        return sj.pow(base, self)

    def __neg__(self):
        return sj.neg(self)

    # It's dangerous to compare just based on the name attribute
    # def __eq__(self, other):
    #     return self.name == other.name

    def __hash__(self):
        # BEWARE: This might be buggy
        return hash(id(self))

    def dot(self, other):
        return sj.dot(self, other)


class Expression(Var):
    def __init__(self, parent1, parent2, operation, varlist=None):
        """
        Initialize an Expression.

        :param parent1: First parent Expression
        :param parent2: Second parent Expression
        :param operation: Operation to combine the two parents
        :param varlist: List of Var objects
            Must be in the same order in which numbers will be passed in upon
            evaluation or differentiation of the Expression
        """
        super().__init__('f')
        self.parent1 = parent1
        self.parent2 = parent2
        self.parents = [self.parent1, self.parent2]
        self.operation = operation
        if varlist is None:
            self._vars = self._get_parent_vars(self.parent1)
            self._vars += [v for v in self._get_parent_vars(self.parent2) if v not in self._vars]
        else:
            self._vars = varlist
        self.matched_vars = self._match_vars_to_parents()

    def set_vars(self, varlist):
        self._vars = varlist
        self.matched_vars = self._match_vars_to_parents()

    @property
    def vars(self):
        return self._vars

    @vars.setter
    def vars(self, varlist):
        # Call set_vars here
        self.set_vars(varlist)

    def _match_vars_to_parents(self):
        """Matches variables to parent1 and parent2
        :return: list[tuple] -- Mapping of vars -> parent Expression objects
        """
        matched_vars = {}
        for var in self.vars:
            parents_of_var = []
            for parent in self.parents:
                if var in self._get_parent_vars(parent):
                    parents_of_var.append(parent)
            matched_vars[var] = parents_of_var
        return matched_vars

    def eval(self, *args):
        """Evaluate this Expression at the specified point

        TODO: Deal with unary operations

        :param args: tuple of values to evaluate the Expression at
        :return: Result (length depends on dimensionality of co-domain)
        """
        if self.parent2 is None:
            return self._unary_eval(*args)
        else:
            return self._binary_eval(*args)

    def deriv(self, *args, mode='forward', var=None):
        """Differentiate this Expression at the specified point.

        Currently just runs forward mode.

        #TODO: Deal with vector outputs
        #TODO: Reverse mode!

        :param args: tuple -- values to evaluate the Expression at
        :param mode: str -- Whether to run in forward or reverse mode
            (possible options: {'forward', 'reverse', 'auto'})
        :param var: Var -- Variable to take derivative with respect to
            Default: None (gets entire Jacobian)
        :return: tuple(Number) | Number -- Result (length depends on dimensionality of co-domain)
        """
        assert mode in ('forward', 'reverse', 'auto'), f'Invalid model specified: {mode}. ' \
                                                       f'Please choose one of "forward", "reverse", "auto".'
        if mode == 'auto':
            if len(self.vars) > 1:
                mode = 'reverse'
            else:
                mode = 'forward'
        if mode == 'forward':
            if var is None:
                res = []
                for var in self.vars:
                    res.append(self._deriv(var, mode, *args))
                if len(res) == 1:
                    return res[0]
                else:
                    return np.array(res)
            else:
                return self._deriv(var, mode, *args)
        else:
            rev = sj.reverse(self)
            return rev(*args, var=var)

    def _deriv(self, var, mode, *args):
        if self.parent2 is None:
            return self._unary_deriv(*args, mode=mode, var=var)
        else:
            return self._binary_deriv(*args, mode=mode, var=var)

    def _unary_eval(self, *args):
        return self.operation.eval(self._eval_parent(self.parent1, *args))

    def _binary_eval(self, *args):
        p1_args, p2_args = self._parse_args(*args)
        return self.operation.eval(self._eval_parent(self.parent1, *p1_args), self._eval_parent(self.parent2, *p2_args))

    def _unary_deriv(self, *args, mode='forward', var=None):
        res = self.operation.deriv(self._eval_parent(self.parent1, *args),
                                   self._deriv_parent(self.parent1, var, *args))
        return res

    def _binary_deriv(self, *args, mode='forward', var=None):
        p1_args, p2_args = self._parse_args(*args)
        res = self.operation.deriv(self._eval_parent(self.parent1, *p1_args),  # Evaluate and store once
                                   self._deriv_parent(self.parent1, var, *p1_args),
                                   self._eval_parent(self.parent2, *p2_args),
                                   self._deriv_parent(self.parent2, var, *p2_args))
        return res

    def _get_input_args(self, parent, *args):
        """Parse the arguments in terms of the ordering for the parent

        :param parent: Parent Expression to be evaluated
        :param args: Arguments in order of self.vars
        :return: list[Var]
        """
        return get_input_args(parent, self.vars, *args)

    def _check_input_length(self, *args):
        """Check that the input length matches this function's domain dimensionality.

        :param args: Input arguments
        :return: None

        :raises: AssertionError
        """
        assert len(args) == len(self.vars), \
            f'Input length does not match dimension of Expression domain ({len(args)}, {len(self.vars)})'

    def _parse_args(self, *args):
        """Check input length and parse arguments in correct order for each parent.

        :param args: tuple[Numeric] -- Arguments to be parsed
        :return: (tuple[Numeric], tuple[Numeric]) -- p1_args and p2_args
        """
        self._check_input_length(*args)
        p1_args = self._get_input_args(self.parent1, *args)
        p2_args = self._get_input_args(self.parent2, *args)
        return p1_args, p2_args

    @staticmethod
    def _get_parent_vars(parent: Union[Var, Number, None]):
        """Get the vars for given parent

        :param parent: Var | Number | None -- The parent of interest
        :return: list[Var]
        """
        if parent is None or isinstance(parent, Number):
            return []
        else:
            return parent.vars[:]

    @staticmethod
    def _eval_parent(parent: Union[Var, Number], *args) -> Number:
        """Evaluate a parent at `args`, checking if the parent is None or a Number

        :param parent: Var | Number | None -- The parent to evaluate
        :return: Number
        """
        if not isinstance(parent, Var):
            return parent
        else:
            return parent(*args)

    @staticmethod
    def _deriv_parent(parent: Union[Var, Number], var: Var, *args) -> Number:
        """Evaluate derivative of a parent at `args`, checking if the parent is a Number

        :param parent: The parent of interest
        :param var: Var -- Variable with respect to which the derivative is
            being taken
        :param args: Point to differentiate parent at
        :return: Number
        """
        if isinstance(parent, Var):
            if var in parent.vars:
                return parent.deriv(*args, var=var)
        return 0

    def __call__(self, *args, **kwargs):
        return self.eval(*args)

    def __str__(self):
        # TODO: Make this more informative
        if self.parent2 is None:
            return self.operation.opstr(self.parent1)
        else:
            return self.operation.opstr(self.parent1, self.parent2)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __hash__(self):
        # BEWARE: This might be buggy
        # For some reason this isn't inherited from Var...
        return hash(id(self))


class VectorExpression:
    """
    A wrapper expression that can handle vector-valued outputs
    """
    def __init__(self, expressions, varlist):
        self._vars = varlist
        self._expressions = self._match_vars_to_expressions(varlist, expressions)

    @property
    def vars(self):
        return self._vars

    @vars.setter
    def vars(self, varlist):
        self._vars = varlist
        self._expressions = self._match_vars_to_expressions(varlist, self._expressions.keys())  # This might not work

    def eval(self, *args):
        return [e(*self._get_expr_args(e, *args)) for e in self._expressions]

    def deriv(self, *args, mode='forward', var=None):
        res = np.zeros((len(self._expressions), len(self._vars)))
        for i, (e, v) in enumerate(self._expressions.items()):
            expr_args = self._get_expr_args(e, *args)
            expr_deriv = e.deriv(*expr_args, mode=mode)
            res[i, :] = self._parse_results(expr_deriv, v)
        return res

    def _get_expr_args(self, expr, *args):
        expr_vars_idx = self._expressions.get(expr, [])
        res = []
        for idx in expr_vars_idx:
            res.append(args[idx])
        return res

    def _parse_results(self, result_set, expr_vars):
        res = np.zeros(len(self._vars))
        res[expr_vars] = result_set
        return res

    @staticmethod
    def _match_vars_to_expressions(varlist, expressions):
        return {expr: VectorExpression._get_var_order(varlist, expr) for expr in expressions}

    @staticmethod
    def _get_var_order(varlist, expr):
        if not isinstance(expr, Var):
            return []
        var_order = [varlist.index(var) for var in expr.vars]
        return var_order

    def __call__(self, *args, **kwargs):
        return self.eval(*args)

    def __repr__(self):
        return '(' + ', '.join([str(e) for e in self._expressions.keys()]) + ')'

    def __str__(self):
        return repr(self)

def get_input_args(expression, varlist, *args):
    """Parse the arguments in terms of the ordering for the parent

    :param expression: Expression to be evaluated
    :param varlist: Ordered list of variables that matches `args`
    :param args: Arguments in order of self.vars
    :return: list[Var]
    """
    if not isinstance(expression, Var):
        return []
    input_args = [args[varlist.index(parent_var)] for parent_var in expression.vars]
    return input_args



from abc import ABC

import numpy as np
from numbers import Number
from .expression import Var, Expression


class OperationType(type):
    def __str__(self):
        return self.__name__


class BaseOperation:
    __metaclass__ = OperationType

    @classmethod
    def check_type(cls, *args):
        """
        :param args: tuple(Object) -- Parameters to check
        :return: None
        :raises: AssertionError if all elements of args are not a Var or Number
        """
        for x in args:
            if not isinstance(x, Var) and not isinstance(x, Number):
                raise TypeError("Not a number/Variable/Expression")

    @classmethod
    def reverse(cls, *args):
        """
        :param args: numbers -- values of the parent nodes
        :return: numbers -- the reverse mode derivative at the current node
        """
        raise NotImplementedError()


class UnaryOperation(BaseOperation, ABC):
    @classmethod
    def expr(cls, expr):
        """Create a new expression

        :param expr: Var | Number -- Parent expression
        :return: Var | Number -- new expression
        """
        cls.check_type(expr)
        return Expression(expr, None, cls)

    @classmethod
    def eval(cls, num):
        """Evaluate the operation at `num`

        :param num: Number -- argument to the unary operation
        :return: Number -- result of evaluation
        """
        raise NotImplementedError()

    @classmethod
    def deriv(cls, val, der):
        """Evaluate the derivative at value = `val` and derivative = `deriv`

        :param val: Number -- Value of the argument
        :param der: Number -- Value of the derivative of the argument
        :return: Number -- The derivative
        """
        raise NotImplementedError()

    @classmethod
    def opstr(cls, expr):
        """For use in the __str__ method of Expression

        :param expr: Expression -- Argument to operation
        :return: str
        """
        raise NotImplementedError()


class BinaryOperation(BaseOperation, ABC):
    @classmethod
    def expr(cls, expr1, expr2):
        """Create a new expression

        :param expr1: Var | Number -- Expression or number to become parent 1
        :param expr2: Var | Number -- Expression or number to become parent 2
        :return:
        """
        cls.check_type(expr1, expr2)
        return Expression(expr1, expr2, cls)

    @classmethod
    def eval(cls, num1, num2):
        """Evaluate the expression at `num2`, `num2`

        :param num1: Number -- First argument to the operation
        :param num2: Number -- Second argument to the operation
        :return:
        """
        raise NotImplementedError()

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        """Differentiate the expression at the given values

        :param num1: Number -- Value of parent 1
        :param deriv1: Number -- Value of derivative of parent 1
        :param num2: Number -- Value of parent 2
        :param deriv2: Number -- Value of parent 2
        :return:
        """
        raise NotImplementedError()

    @classmethod
    def opstr(cls, expr1, expr2):
        """For use in the __str__ method of Expression

        :param expr1: Expression -- First argument to operation
        :param expr2: Expression -- Second argument to operation
        :return: str
        """
        raise NotImplementedError()


class Add(BinaryOperation):
    @classmethod
    def eval(cls, num1, num2):
        # super().check_type(num1)
        # super().check_type(num2)
        return num1 + num2

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        return deriv1 + deriv2

    @classmethod
    def reverse(cls, *args):
        return (1, 1)

    @classmethod
    def opstr(cls, expr1, expr2):
        return f'{str(expr1)} + {str(expr2)}'


class Sub(BinaryOperation):
    @classmethod
    def eval(cls, num1, num2):
        return num1 - num2

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        return deriv1 - deriv2

    @classmethod
    def reverse(cls, *args):
        return (1, -1)

    @classmethod
    def opstr(cls, expr1, expr2):
        return f'{str(expr1)} - {str(expr2)}'


class Mul(BinaryOperation):
    @classmethod
    def eval(cls, num1, num2):
        return num1 * num2

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        return num1 * deriv2 + num2 * deriv1

    @classmethod
    def reverse(cls, *args):
        return (args[1], args[0])

    @classmethod
    def opstr(cls, expr1, expr2):
        return f'{str(expr1)} * {str(expr2)}'


class Div(BinaryOperation):
    @classmethod
    def eval(cls, num1, num2):
        return num1 / num2

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        return -(num1 * deriv2 - num2 * deriv1) / (num2 ** 2)

    @classmethod
    def reverse(cls, *args):
        return (1 / args[1], - args[0] / args[1]**2)

    @classmethod
    def opstr(cls, expr1, expr2):
        return f'{str(expr1)} / {str(expr2)}'


class Pow(BinaryOperation):
    @classmethod
    def eval(cls, num1, num2):
        return num1 ** num2

    @classmethod
    def deriv(cls, num1, deriv1, num2, deriv2):
        if num1 > 0:
            result = np.exp(num2 * np.log(num1)) * (deriv2 * np.log(num1) + num2 * deriv1 / num1)
            return result

        # deal with the case when log(num1) is complex
        else:
            result = np.exp(num2 * np.log(num1 + 0j)) * (deriv2 * np.log(num1 + 0j) + num2 * deriv1 / num1)
            return np.real(result)

    @classmethod
    def reverse(cls, *args):
        a = args[0] # parent 1 value -- base
        b = args[1] # parent 2 value -- exponent
        if a < 0:
            return np.real((b * a ** (b-1), np.log(a+0j) * a ** b))
        else:
            return (b * a ** (b-1), np.log(a) * a ** b)

    @classmethod
    def opstr(cls, expr1, expr2):
        return f'{str(expr1)}^{str(expr2)}'


class Sqrt(UnaryOperation):
    @classmethod
    def eval(cls, num1):
        return np.sqrt(num1)

    @classmethod
    def deriv(cls, num1, deriv1):
        return 1 / 2 / np.sqrt(num1) * deriv1

    @classmethod
    def reverse(cls, *args):
        a = args[0]
        return 1 / 2 / np.sqrt(a)

    @classmethod
    def opstr(cls, expr):
        return f'sqrt({str(expr)})'
    

class Neg(UnaryOperation):
    @classmethod
    def eval(cls, num1):
        return -num1

    @classmethod
    def deriv(cls, num1, deriv1):
        return -deriv1

    @classmethod
    def reverse(cls, *args):
        return -1

    @classmethod
    def opstr(cls, expr):
        return f'-{str(expr)}'

    
class Exp(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.exp(num)

    @classmethod
    def deriv(cls, num, deriv):
        return deriv*np.exp(num)

    @classmethod
    def reverse(cls, *args):
        return np.exp(args[0])

    @classmethod
    def opstr(cls, expr):
        return f'exp({str(expr)})'


class NLog(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.log(num)

    @classmethod
    def deriv(cls, val, der):
        return der / val

    @classmethod
    def reverse(cls, *args):
        return 1 / args[0]

    @classmethod
    def opstr(cls, expr):
        return f'ln({str(expr)})'


class Log(BinaryOperation):
    @classmethod
    def eval(cls, num, base=np.e):
        return np.log(num) / np.log(base)

    @classmethod
    def deriv(cls, val, der, base=np.e, base_der=0):
        return (((der / val) * np.log(base)) - ((base_der / base) * np.log(val))) / (np.log(base)**2)

    @classmethod
    def reverse(cls, *args):
        a = args[0]
        b = args[1]
        return (1 / np.log(b) / a, - np.log(a) / np.log(b)**2 / b)

    @classmethod
    def opstr(cls, expr1, expr2):
        # This might be wrong
        return f'log_{str(expr2)}({str(expr1)})'


class Sin(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.sin(num)

    @classmethod
    def deriv(cls, val, der):
        return np.cos(val) * der

    @classmethod
    def reverse(cls, *args):
        return np.cos(args[0])

    @classmethod
    def opstr(cls, expr):
        return f'sin({str(expr)})'


class Cos(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.cos(num)

    @classmethod
    def deriv(cls, val, der):
        return -np.sin(val) * der

    @classmethod
    def reverse(cls, *args):
        return -np.sin(args[0])

    @classmethod
    def opstr(cls, expr):
        return f'cos({str(expr)})'


class Tan(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.tan(num)

    @classmethod
    def deriv(cls, val, der):
        return der / (np.cos(val) ** 2)

    @classmethod
    def reverse(cls, *args):
        return 1 / (np.cos(args[0]) ** 2)

    @classmethod
    def opstr(cls, expr):
        return f'tan({str(expr)})'


class Csc(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return 1/np.sin(num)

    @classmethod
    def deriv(cls, val, der):
        return -der*(1/np.sin(val))*(1/np.tan(val))

    @classmethod
    def reverse(cls, *args):
        return -1*(1/np.sin(args[0]))*(1/np.tan(args[0]))

    @classmethod
    def opstr(cls, expr):
        return f'csc({str(expr)})'


class Sec(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return 1/np.cos(num)

    @classmethod
    def deriv(cls, val, der):
        return der*(1/np.cos(val))*np.tan(val)

    @classmethod
    def reverse(cls, *args):
        return 1*(1/np.cos(args[0]))*np.tan(args[0])

    @classmethod
    def opstr(cls, expr):
        return f'sec({str(expr)})'


class Cot(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return 1/np.tan(num)

    @classmethod
    def deriv(cls, val, der):
        return -der*(1/np.sin(val))**2

    @classmethod
    def reverse(cls, *args):
        return -1*(1/np.sin(args[0]))**2

    @classmethod
    def opstr(cls, expr):
        return f'cot({str(expr)})'


class ArcSin(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.arcsin(num)

    @classmethod
    def deriv(cls, val, der):
        return der / np.sqrt(1 - val**2)

    @classmethod
    def reverse(cls, *args):
        return 1 / np.sqrt(1 - args[0]**2)


class ArcCos(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.arccos(num)

    @classmethod
    def deriv(cls, val, der):
        return - der / np.sqrt(1 - val**2)

    @classmethod
    def reverse(cls, *args):
        return - 1 / np.sqrt(1 - args[0]**2)


class ArcTan(UnaryOperation):
    @classmethod
    def eval(cls, num):
        return np.arctan(num)

    @classmethod
    def deriv(cls, val, der):
        return der / (1 + val**2)

    @classmethod
    def reverse(cls, *args):
        return 1 / (1 + args[0]**2)


# Vector operations
##class Dot(BinaryOperation):
##    @classmethod
##    def eval(cls, vec1, vec2):
##        return np.dot(vec1, vec2)
##
##    @classmethod
##    def deriv(cls, vec1, deriv1, vec2, deriv2):
##        return np.dot(vec1, deriv2) + np.dot(deriv1, vec2)



"""
Implementing reverse mode differentiation
"""
from numbers import Number

import numpy as np

from superjacob.expression import Expression, get_input_args


class ReverseDiff:
    """
    Class implementing reverse mode differentiation
    """

    def __init__(self, expr):
        self.expr = expr
        self.vars = expr.vars
        self.trace = []

    def forward(self, expr, *args, child=None):
        """Compute the forward pass of reverse mode differentiation

        :param expr: Var -- Expression to be differentiated
        :param args: tuple -- Arguments to differentiate the expression at
        :param child: Expression -- The last node visited in the trace (child of current expr)
        :return: Number -- evaluated expression
        """
        if expr is None:
            return None
        if isinstance(expr, Number):
            return expr
        if expr in self.trace:  # Here we check if we have already visited this node
            self._add_child(expr, child)
            return args[0]
        if not isinstance(expr, Expression):
            node = TraceNode(expr, args[0], [1])
            node.add_child(child)
            self.trace.append(node)
            return args[0]
        else:
            # Parsing args
            p1_args, p2_args = get_input_args(expr.parent1, expr.vars, *args), \
                               get_input_args(expr.parent2, expr.vars, *args)

            node = TraceNode(expr, child=child)

            parvals = self.forward(expr.parent1, *p1_args, child=node), \
                      self.forward(expr.parent2, *p2_args, child=node)

            # For now, implementing unary/binary as an if-statement. Consider subclassing
            if parvals[1] is None:
                currval = expr.operation.eval(parvals[0])
                d1val = expr.operation.reverse(parvals[0])
                derivs = [d1val]
            else:
                currval = expr.operation.eval(*parvals)
                derivs = expr.operation.reverse(*parvals)
            node.currval = currval
            node.derivs = derivs

            self.trace.append(node)
            return currval

    def reverse(self, var=None):
        """Compute the reverse pass of forward mode differentiation

        :param var: Var -- The variable with respect to which the derivative is taken
        :return: gradient
        """
        res = np.zeros(len(self.vars))
        for node in self.trace[::-1]:
            node_bar = node.bar
            if node in self.vars:
                if var and node == var:
                    return node_bar
                res[self.vars.index(node)] = node_bar
        return res

    def _add_child(self, parent, child):
        if child:
            self.trace[self.trace.index(parent)].add_child(child)

    def __call__(self, *args, **kwargs):
        self.forward(self.expr, *args)
        return self.reverse(**kwargs)


class TraceNode:
    """
    A lightweight class for storing elements of the evaluation trace.
    """

    def __init__(self, expr, currval=None, derivs=[], child=None):
        self.expr = expr
        self.currval = currval
        self.derivs = derivs
        if child is None:
            self.children = []
        else:
            self.children = [child]
        self._bar = None

    def add_child(self, child):
        if child:
            self.children.append(child)

    @property
    def bar(self):
        if self._bar:
            return self._bar
        if not self.children:
            self._bar = 1
        else:
            res = 0
            for child in self.children:
                res += child.bar * child.deriv_parent(self.expr)
            self._bar = res
        return self._bar

    @bar.setter
    def bar(self, value):
        assert not self._bar, 'Bar already set'
        self._bar = value

    def deriv_parent(self, parent_expr):
        """Get the derivative of this node with respect to `parent_expr`

        :param parent_expr: TraceNode -- parent with respect to which the derivative should be taken
        :return: Number
        """
        return self.derivs[self.expr.parents.index(parent_expr)]

    def __eq__(self, other):
        if isinstance(other, TraceNode):
            return self.expr == other.expr
        else:
            return self.expr == other

    def __str__(self):
        return str(self.expr)

    def __repr__(self):
        return repr(self.expr)


import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()


setuptools.setup(
    name='superjacob',
    version='0.0.1',
    author='Benjamin Levy, Mark Miao, Zhenru Wang, Kidus Negesse',
    description='Yet another autodiff package (YAAP',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ADMonsters/cs207-FinalProject',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.5',
)
